from pydantic import BaseModel, HttpUrl, Field, validator
from typing import Union, Dict, Optional, Literal, List
from datetime import datetime
from .links import APLink, APMention
from .collections import (
    APCollection,
    APOrderedCollection,
    APOrderedCollectionPage,
    APCollectionPage
)
from .objects import (
    APObject,
    APEvent,
    APPlace,
    APRelationship,
    APProfile,
    APTombstone,
    APArticle,
    APAudio,
    APDocument,
    APImage,
    APNote,
    APPage,
    APVideo
    )

class AnyCollection(BaseModel):
    __root__: Union[APCollection, APOrderedCollection]

# AnyAPObject is a union of many APObject types, including AnyCollection
class AnyAPObject(BaseModel):
    __root__: Union[
        APObject,
        APArticle,
        APAudio,
        APDocument,
        APEvent,
        APImage,
        APNote,
        APPage,
        APPlace,
        APProfile,
        APRelationship,
        APTombstone,
        APVideo,
        AnyCollection
    ]
class DateTime(BaseModel):
    __root__: Union[str | datetime]

class LanguageTag(BaseModel):
    __root__: str

class MediaType(BaseModel):
    __root__: str

    @validator("__root__")
    def validate_media_type(cls, value):
        """
        Validation helper for MIME types.
        This validates the type of media file in use
        it recognizes "image/jpeg", "image/png", "application/json", "text/html"

        """
        allowed_mimes = ["image/jpeg", "image/png", "application/json", "text/html"]
        if value and value not in allowed_mimes:
            raise ValueError(f"Invalid MIME type: {value}")
        return value

class Duration(BaseModel):
    __root__: str

class ContextField(BaseModel):
    __root__: Union[str, Dict[str, str]] = Field(alias="@context")

class UrlField(BaseModel):
    __root__: Union[str, APLink]

class IdField(BaseModel):
    __root__: str

class TypeField(BaseModel):
    __root__: str

class DurationField(BaseModel):
    __root__: Duration

class ContentField(BaseModel):
    __root__: str

class ContentMapField(BaseModel):
    __root__: Dict[str, ContentField]

class MediaTypeField(BaseModel):
    __root__: MediaType

class NameField(BaseModel):
    __root__: str

class NameMapField(BaseModel):
    __root__: Dict[str, NameField]

class SummaryField(BaseModel):
    __root__: str

class SummaryMapField(BaseModel):
    __root__: Dict[str, SummaryField]

class EndTimeField(BaseModel):
    __root__: DateTime

class StartTimeField(BaseModel):
    __root__: DateTime

class PublishedField(BaseModel):
    __root__: DateTime

class UpdatedField(BaseModel):
    __root__: DateTime

class RepliesField(BaseModel):
    __root__: Union[str, AnyCollection]

class RelationshipField(BaseModel):
    __root__: Union[str, AnyAPObject]

class IconField(BaseModel):
    __root__: Union[str, APImage, APLink]

class ImageField(BaseModel):
    __root__: Union[str, APImage, APLink]

class AttachmentField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class AudienceField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class InReplyToField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class LocationField(BaseModel): 
    __root__: Union[str, APPlace, APLink]

class PreviewField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class ToField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class BtoField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class CcField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class BccField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class GeneratorField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class ObjectField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class AttributedToField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink, APMention]

class TagField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink, APMention]

class ActorField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class TargetField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class ResultField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class OriginField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class InstrumentField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class OneOfField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class AnyOfField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class SubjectField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink]

class AccuracyField(BaseModel):
    __root__: float

class AltitudeField(BaseModel):
    __root__: float

class LatitudeField(BaseModel):
    __root__: float

class LongitudeField(BaseModel):
    __root__: float

class RadiusField(BaseModel):
    __root__: float = Field(..., ge=0.0)

class UnitsField(BaseModel):
    __root__: Union[Literal['cm', 'feet', 'inches', 'km', 'm', 'miles'], str]

class ClosedField(BaseModel):
    __root__: Union[str, AnyAPObject, APLink, DateTime, bool]

class FormerTypeField(BaseModel):
    __root__: str

class DeletedField(BaseModel):
    __root__: DateTime

class DescribesField(BaseModel):
    __root__: Union[str, AnyAPObject]

class HrefField(BaseModel):
    __root__: HttpUrl


class HreflangField(BaseModel):
    __root__: LanguageTag

class RelField(BaseModel):
    __root__: List[str]

class HeightField(BaseModel):
    __root__: int = Field(..., ge=0)

class WidthField(BaseModel):
    __root__: int = Field(..., ge=0)

class TotalItemsField(BaseModel):
    __root__: int = Field(..., ge=0)

class CollectionCurrentField(BaseModel):
    __root__: Union[str, APCollectionPage, APLink]

class CollectionFirstField(BaseModel):
    __root__: Union[str, APCollectionPage, APLink]

class CollectionLastField(BaseModel):
    __root__: Union[str, APCollectionPage, APLink]


class CollectionItemsField(BaseModel):
    __root__: Union[str, APCollectionPage, APLink]


class CollectionPagePartOfField(BaseModel):
    __root__: Union[str, APCollection, APLink]


class CollectionPageNextField(BaseModel):
    __root__: Union[str, APCollection, APLink]


class CollectionPagePrevField(BaseModel):
    __root__: Union[str, APCollection, APLink]


class OrderedCollectionCurrentField(BaseModel):
    __root__: Union[str, APOrderedCollectionPage, APLink]


class OrderedCollectionFirstField(BaseModel):
    __root__: Union[str, APOrderedCollectionPage, APLink]


class OrderedCollectionLastField(BaseModel):
    __root__: Union[str, APOrderedCollectionPage, APLink]


class OrderedCollectionItemsField(BaseModel):
    __root__: Union[str, APOrderedCollectionPage, APLink]


class OrderedCollectionPagePartOfField(BaseModel):
    __root__: Union[str, APOrderedCollection, APLink]


class OrderedCollectionPageNextField(BaseModel):
    __root__: Union[str, APOrderedCollection, APLink]


class OrderedCollectionPagePrevField(BaseModel):
    __root__: Union[str, APOrderedCollection, APLink]


class StartIndexField(BaseModel):
    __root__: int = Field(..., ge=0)

class SourceField(BaseModel):
    content: str
    media_type: MediaType

class InboxField(BaseModel):
    __root__: Union[str, APOrderedCollection, APLink]

class OutboxField(BaseModel):
    __root__: Union[str, APOrderedCollection, APLink]

class FollowingField(BaseModel):
    __root__: Union[str, APCollection, APOrderedCollection, APLink]

class FollowersField(BaseModel):
    __root__: Union[str, APCollection, APOrderedCollection, APLink]

class LikedField(BaseModel):
    __root__: Union[str, APCollection, APOrderedCollection, APLink]

class LikesField(BaseModel):
    __root__: Union[str, APCollection, APOrderedCollection, APLink]


class SharesField(BaseModel):
    __root__: Union[str, APCollection, APOrderedCollection, APLink]

class StreamsField(BaseModel):
    __root__: Union[str, APCollection, APOrderedCollection, APLink]

class PreferredUsernameField(BaseModel):
    __root__: str


class PreferredUsernameMapField(BaseModel):
    __root__: Dict[str, str]


class ProxyUrlField(BaseModel):
    __root__: str


class OauthAuthorizationEndpointField(BaseModel):
    __root__: str


class OauthTokenEndpointField(BaseModel):
    __root__: str


class ProvideClientKeyField(BaseModel):
    __root__: str


class SignClientKeyField(BaseModel):
    __root__: str


class SharedInboxField(BaseModel):
    __root__: str

class EndpointsField(BaseModel):
    proxy_url: Optional[ProxyUrlField] = None
    oauth_authorization_endpoint: Optional[OauthAuthorizationEndpointField] = None
    oauth_token_endpoint: Optional[OauthTokenEndpointField] = None
    provide_client_key: Optional[ProvideClientKeyField] = None
    sign_client_key: Optional[SignClientKeyField] = None
    shared_inbox: Optional[SharedInboxField] = None
