from __future__ import annotations
from pydantic import HttpUrl, Field, AnyUrl, PositiveInt, validator
from typing import Optional
from serializers.json_serializer import ActivityPubModel

from .fields import (
    UrlField,
    IconField,
    ImageField,
    AttachmentField,
    AudienceField,
    InReplyToField,
    LocationField,
    PreviewField,
    ToField,
    BtoField,
    CcField,
    BccField,
    AttributedToField,
    TagField,
    GeneratorField,
    IdField,
    TypeField,
    ContentField,
    ContentMapField,
    MediaTypeField,
    NameField,
    NameMapField,
    EndTimeField,
    StartTimeField,
    PublishedField,
    RepliesField,
    SummaryField,
    SummaryMapField,
    UpdatedField,
    DurationField, 
    ContextField,
    AccuracyField,
    AltitudeField,
    LatitudeField,
    LongitudeField,
    RadiusField,
    UnitsField,
    DescribesField,
    SubjectField,
    ObjectField,
    RelationshipField,
    FormerTypeField,
    DeletedField,
    SourceField, 
    LikesField, 
    SharesField, 
)

class _APObjectBaseModel(ActivityPubModel):
    """
    Base class for all objects in the ActivityStreams vocabulary

    """
    id: Optional[IdField] = None 
    type: TypeField
    context: Optional[ContextField] = Field(alias="@context")
    name: Optional[NameField] = None

# class Image(_MediaObject):
#     pass

class APObject(_APObjectBaseModel):
    """
    The Object class represents any object in the vocabulary.
    https://www.w3.org/TR/activitystreams-core/#object
    https://www.w3.org/TR/activitystreams-vocabulary/#object-types

    """
    attachment: Optional[AttachmentField] = None
    attributed_to: Optional[AttributedToField] = None
    audience: Optional[AudienceField] = None
    bcc: Optional[BccField] = None
    bto: Optional[BtoField] = None
    cc: Optional[CcField] = None
    content: Optional[ContentField] = None
    content_map: Optional[ContentMapField] = None 
    generator: Optional[GeneratorField] = None
    icon: Optional[IconField] = None
    image: Optional[ImageField] = None
    in_reply_to: Optional[InReplyToField] = None
    location: Optional[LocationField] = None
    end_time: Optional[EndTimeField] = None
    preview: Optional[PreviewField] = None
    published: Optional[PublishedField] = None
    replies: Optional[RepliesField] = None
    start_time: Optional[StartTimeField] = None
    summary: Optional[SummaryField] = None
    summary_map: Optional[SummaryMapField] = None
    name: Optional[NameField] = None
    name_map: Optional[NameMapField] = None
    tag: Optional[TagField] = None
    updated: Optional[UpdatedField] = None
    url: Optional[UrlField] = None
    to: Optional[ToField] = None
    duration: Optional[DurationField] = None
    media_type: Optional[MediaTypeField] = None

    # added these field based on the typescript definitions, 
    # TODO yet to verify if this should be included or not
    source: Optional[SourceField] = None
    likes: Optional[LikesField] = None
    shares: Optional[SharesField] = None


    @validator("end_time", "start_time")
    def validate_time_constraints(cls, v, values, field):
        if field.name == "end_time" and v and values.get("start_time") and v < values["start_time"]:
            raise ValueError("end_time cannot be before start_time")
        return v

class APEvent(APObject):
    """
    Represents any kind of event.

    """
    type: TypeField = Field("Event", const=True)

class APPlace(APObject):
    """
    Represents a logical or physical location. 
    https://www.w3.org/TR/activitystreams-vocabulary/#dfn-place
    https://www.w3.org/TR/activitystreams-vocabulary/#places

    """
    type: TypeField = Field("Place", const=True)
    accuracy: AccuracyField
    altitude: AltitudeField
    latitude: LatitudeField
    logitude: LongitudeField
    raduis: RadiusField
    units: UnitsField

    @validator("units")
    def validate_unit(cls, value):
        allowed_units = ["cm", " feet", " inches", " km", " m", " miles" ]
        if value and value not in allowed_units:
            raise ValueError(f"Invalid units type: {value}")
        return value

class APProfile(APObject):
    """
    Represents Profile
    """
    type: TypeField = Field("Profile", const=True)
    describes: DescribesField

class APRelationship(APObject):
    type: TypeField = Field("Relationship", const=True)
    subject: SubjectField
    object: ObjectField
    relationship: RelationshipField

class APTombstone(APObject):
    type: TypeField = Field("Tombstone", const=True)
    former_type: FormerTypeField
    deleted: DeletedField

class APArticle(APObject):
    type: TypeField = Field("Article", const=True)

class APAudio(APObject):
    type: TypeField = Field("Audio", const=True)

class APDocument(APObject):
    type: TypeField = Field("Document", const=True)

class APImage(APObject):
    type: TypeField = Field("Image", const=True)

class APNote(APObject):
    type: TypeField = Field("Note", const=True)

class APPage(APObject):
    type: TypeField = Field("Page", const=True)

class APVideo(APObject):
    type: TypeField = Field("Video", const=True)