import warnings
from pydantic import BaseModel, root_validator
from typing import Dict, Union, Generator, Callable, Any

def to_camel_case(snake_str: str) -> str:
    """
    Converts snake_case to camelCase
    """
    pascal_string = "".join(x.capitalize() for x in snake_str.lower().split("_"))
    return snake_str[0].lower() + pascal_string[1:]

class ActivityPubModel(BaseModel):
    """
    A robust class combining serialization, deserialization, and object type handling
    for ActivityPub models.
    """

    _subtypes_: Dict[str, type] = dict()

    class Config:
        alias_generator = to_camel_case

    @root_validator(pre=True)
    @classmethod
    def set_object_type(cls, values: Dict[str, str]) -> Dict[str, str]:
        """
        Reads the type of an object from its class and adds the class to the value 
        so that it appears in the serialized representation.
        """
        values["type"] = cls.__name__
        return values

    def __init_subclass__(cls) -> None:
        if cls.__name__ in cls._subtypes_:
            warnings.warn(
                f"A subclass called {cls.__name__} is already initialized. Unexpected behavior might occur."
            )
        cls._subtypes_[cls.__name__] = cls

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., None], None, None]:
        yield cls._convert_to_real_type

    @classmethod
    def _convert_to_real_type(cls, data: Union[str, Dict[str, Any]]) -> Any:
        """
        Ascertains the data type of a deserialized object. This ensures the correct object is created.
        """
        if isinstance(data, str):
            raise NotImplementedError("Actor needs dereferencing, not yet supported")

        if isinstance(data, dict):
            data_type = data.get("type")
            if data_type is None:
                raise ValueError("Missing 'type' in ActivityPub Model")

            sub = cls._subtypes_.get(data_type)
            if sub is None:
                raise TypeError(f"Unsupported sub-type: {data_type}")

            return sub(**data)

    @classmethod
    def parse_obj(cls, obj: Union[str, Dict[str, Any]]) -> Any:
        """
        Deserializes a string into a Python object.
        """
        return cls._convert_to_real_type(obj)

    def dict(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """
        Converts an object to a Python dict with by_alias=True and exclude_none=True by default.
        """
        if "by_alias" not in kwargs:
            kwargs.update(by_alias=True)
        if "exclude_none" not in kwargs:
            kwargs.update(exclude_none=True)
        return super().dict(*args, **kwargs)

    @classmethod
    def serialize(cls, obj: BaseModel) -> str:
        """
        Serializes an object into a JSON string.
        """
        return obj.json(indent=4)

    @classmethod
    def deserialize(cls, json_str: str, model: BaseModel) -> BaseModel:
        """
        Deserializes a JSON string into the appropriate model object.
        """
        return model.parse_raw(json_str)
